/* configEncoder.c */

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stddef.h>
#include "GhsControlStructs.h"
#include "nomenclature.h"

#ifdef _WIN32
#define NRF_LOG_DEBUG printf
#include <windows.h>
#else
#include "nrf_log.h"
#endif

// From btle_utils.c
bool hexToLittleEndianByte(char* hexString, unsigned char* byteArray);
int twoByteEncode(unsigned char* msmtBuf, int index, unsigned short value);
int fourByteEncode(unsigned char* msmtBuf, int index, unsigned long value);

#if (USES_SYSTEM_AVAS == 1 || USES_HEADERS_AVAS == 1 || USES_MSMT_AVAS == 1)
static void cleanUpAvas(s_Avas** avas, unsigned short numberOfAvas)
{
    if (numberOfAvas > 0 && avas != NULL)
    {
        int i;
        for (i = 0; i < numberOfAvas; i++)
        {
            if (avas[i] != NULL)
            {
                if (avas[i]->value != NULL)
                {
                    free(avas[i]->value);
                }
                free(avas[i]);
            }
        }
        free(avas);
        avas = NULL;
    }
}

static bool allocateAvas(s_Avas*** avasPtr, unsigned short numberOfAvas)
{
    s_Avas** avas = *avasPtr;
    if (avas != NULL)
    {
        free(avas);
    }
    avas = (s_Avas**)calloc(1, numberOfAvas * sizeof(s_Avas*));
    if (avas == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for %d AVA pointers.", numberOfAvas);
        *avasPtr = NULL;
        return false;
    }
    *avasPtr = avas;
    return true;
}

static bool addAva(s_Avas** avasPtr, s_Avas* avaIn)
{
    s_Avas* avas = *avasPtr;
    avas = (s_Avas*)calloc(1, sizeof(s_Avas));
    if (avas == NULL)
    {
        NRF_LOG_DEBUG("Could not allocate memory for the Ava.");
        return false;
    }
    avas->value = (unsigned char*)calloc(1, avaIn->length * sizeof(unsigned char));
    if (avas->value == NULL)
    {
        free(avas);
        *avasPtr = NULL;
        NRF_LOG_DEBUG("Could not allocate memory for the Ava value field of %d bytes.", avaIn->length);
        return false;
    }
    avas->attrId = avaIn->attrId;
    avas->length = avaIn->length;
    memcpy(avas->value, avaIn->value, avaIn->length);
    *avasPtr = avas; // put the allocated value back into the passed pointer to pointer.
    return true;
}
#endif

static bool checkSystemInfo(s_SystemInfo *systemInfo)
{
    if (systemInfo == NULL)
    {
        NRF_LOG_DEBUG("SystemInfo not initialized. Need to call initializeSystemInfo");
        return false;
    }
    return true;
}

static bool checkGhsMsmt(s_GhsMsmt *ghsMsmt)
{
    if (ghsMsmt == NULL)
    {
        NRF_LOG_DEBUG("s_GhsMsmt has not been initialized. Call one of the create measurement methods");
        return false;
    }
    return true;
}

static bool checkMsmtGroup(s_MsmtGroup *msmtGroup)
{
    if (msmtGroup == NULL)
    {
        NRF_LOG_DEBUG("s_MsmtGroup has not been initialized. Call initializeMsmtGroup()");
        return false;
    }
    return true;
}

static bool checkMsmtGroupData(s_MsmtGroupData* msmtGroupData)
{
    if (msmtGroupData == NULL)
    {
        NRF_LOG_DEBUG("s_MsmtGroupData has not been generated. Call createMsmtGroupDataArray()");
        return false;
    }
    return true;
}

//======================= Time Info
void cleanUpTimeInfo(s_TimeInfo** timeInfoPtr)
{
    s_TimeInfo* timeInfo = *timeInfoPtr;
    if (timeInfo != NULL)
    {
        free(timeInfo);
        *timeInfoPtr = NULL;
    }
}

void cleanUpTimeInfoData(s_TimeInfoData** timeInfoDataPtr)
{
    s_TimeInfoData* timeInfoData = *timeInfoDataPtr;
    if (timeInfoData != NULL)
    {
        if (timeInfoData->timeInfoBuf != NULL)
        {
            free(timeInfoData->timeInfoBuf);
        }
        free(timeInfoData);
        *timeInfoDataPtr = NULL;
    }
}

bool createTimeInfo(s_TimeInfo **timeInfoPtr, bool usesAbsTime, bool usesBoTime, bool usesRelativeTime, bool allowSetTime, unsigned long timeResolution)
{
    s_TimeInfo* timeInfo = *timeInfoPtr;
    if (timeInfo != NULL)
    {
        cleanUpTimeInfo(timeInfoPtr);
    }
    timeInfo = (s_TimeInfo *)calloc(1, sizeof(s_TimeInfo));
    if (timeInfo == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for the s_TimeInfo struct");
        return false;
    }
    if (usesAbsTime)
    {
        timeInfo->timeFlagsAbsTime = TIME_FLAGS_ABS_TIME;
    }
    else if (usesBoTime)
    {
        timeInfo->timeFlagsBoTime = TIME_FLAGS_BO_TIME;
    }
    else if (usesRelativeTime)
    {
        timeInfo->timeFlagsRelTime = TIME_FLAGS_REL_TIME;
    }
    if (allowSetTime && (usesAbsTime || usesBoTime))
    {
        timeInfo->timeFlagsSetTime = TIME_FLAGS_SET_TIME;
    }
    timeInfo->timeResolution = timeResolution;
    timeInfo->timeSync = MDC_TIME_SYNC_NONE;
    *timeInfoPtr = timeInfo;
    return true;
}

bool createCurrentTimeDataBuffer(s_TimeInfoData** timeInfoDataPtr, s_TimeInfo *timeInfo)
{

    if (timeInfo == NULL)
    {
        NRF_LOG_DEBUG("TimeInfo struct has not been initialized");
        return false;
    }
    unsigned short index = 0;
    bool supportsCurrentTime = (timeInfo->timeFlagsAbsTime != 0) 
                        || (timeInfo->timeFlagsBoTime != 0) 
                        || (timeInfo->timeFlagsRelTime != 0);
    unsigned short flags = (unsigned short)(timeInfo->timeFlagsAbsTime | timeInfo->timeFlagsBoTime | timeInfo->timeFlagsRelTime | timeInfo->timeFlagsSetTime |
                                timeInfo->timeFlagsTimeFault);
    s_TimeInfoData* timeInfoData = *timeInfoDataPtr;
    if (timeInfoData != NULL)
    {
        cleanUpTimeInfoData(timeInfoDataPtr);
    }
    timeInfoData = (s_TimeInfoData *)calloc(1, sizeof(s_TimeInfoData));
    if (timeInfoData == NULL)
    {
        NRF_LOG_DEBUG("Could not allocate memory for time info Data struct");
        return false;
    }
    unsigned short timeInfoLength = 6;
    timeInfoLength = timeInfoLength + (supportsCurrentTime ? 14 : 0);
    timeInfoData->timeInfoBuf = (unsigned char *)calloc(1, timeInfoLength * sizeof(unsigned char));
    if (timeInfoData->timeInfoBuf == NULL)
    {
        NRF_LOG_DEBUG("Could not allocate memory for time info buffer");
        free(timeInfoData);
        *timeInfoDataPtr = NULL;
        return false;
    }
    timeInfoData->currentTime_index = -1;
    timeInfoData->timeSync_index = -1;
    timeInfoData->dataLength = (unsigned short)timeInfoLength;
    index = twoByteEncode(timeInfoData->timeInfoBuf, index, COMMAND_GET_CURRENT_TIME);
    timeInfoData->flags_index = index;
    index = twoByteEncode(timeInfoData->timeInfoBuf, index, flags);
    index = twoByteEncode(timeInfoData->timeInfoBuf, index, (timeInfoLength - 6));
    if (supportsCurrentTime)
    {
        timeInfoData->currentTime_index = index;
        index = index + 8;      // Place for current time
        index = fourByteEncode(timeInfoData->timeInfoBuf, index, timeInfo->timeResolution);
        timeInfoData->timeSync_index = index;
        index = twoByteEncode(timeInfoData->timeInfoBuf, index, MDC_TIME_SYNC_NONE);
    }
    *timeInfoDataPtr = timeInfoData;
    return true;
}

// ====================== System Info
void cleanUpSystemInfo(s_SystemInfo** systemInfoPtr)
{
    s_SystemInfo* systemInfo = *systemInfoPtr;
    if (systemInfo == NULL) return;
    if (systemInfo->specializations != NULL)
    {
        free(systemInfo->specializations);
    }
    #if (USES_SYSTEM_AVAS == 1)
    cleanUpAvas(systemInfo->avas, systemInfo->numberOfAvas);
    #endif
    free(systemInfo);
    *systemInfoPtr = NULL;
}

void cleanUpSystemInfoData(s_SystemInfoData** systemInfoDataPtr)
{
    s_SystemInfoData* systemInfoData = *systemInfoDataPtr;
    if (systemInfoData != NULL)
    {
        if (systemInfoData->systemInfoBuf != NULL)
        {
            free(systemInfoData->systemInfoBuf);
        }
        free(systemInfoData);
        *systemInfoDataPtr = NULL;
    }
}

bool createSystemInfo(s_SystemInfo **systemInfoPtr, unsigned short numberOfSpecializations)
{
    s_SystemInfo* systemInfo = *systemInfoPtr;
    if (systemInfo != NULL)
    {
        cleanUpSystemInfo(systemInfoPtr);
    }
    systemInfo = (s_SystemInfo *)calloc(1, sizeof(s_SystemInfo));
    if (systemInfo == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for the s_SystemInfo struct");
        return false;
    }
    systemInfo->currentSpecializationCount = 0;
    systemInfo->numberOfSpecializations = numberOfSpecializations;
    systemInfo->specializations = (s_Specialization *)calloc(1, numberOfSpecializations * sizeof(s_Specialization));
    if (systemInfo->specializations == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for the number of specializations %d", numberOfSpecializations);
        free(systemInfo);
        *systemInfoPtr = NULL;
        return false;
    }
    *systemInfoPtr = systemInfo;
    return true;
}

bool addSpecialization(s_SystemInfo **systemInfoPtr, unsigned short specializationTermCode, unsigned short version)
{
    s_SystemInfo* systemInfo = *systemInfoPtr;
    if (!checkSystemInfo(systemInfo)) return false;

    if (systemInfo->currentSpecializationCount >= systemInfo->numberOfSpecializations)
    {
        NRF_LOG_DEBUG("Number of specializations exceeded. Cannot add more. Skipping.");
        return false;
    }
    systemInfo->specializations[systemInfo->currentSpecializationCount].specialization = specializationTermCode;
    systemInfo->specializations[systemInfo->currentSpecializationCount++].version = version;
    return true;
}

bool setRequiredSystemInfoStrings(s_SystemInfo **systemInfoPtr, char *manufacturerName, char *modelNumber)
{
    s_SystemInfo* systemInfo = *systemInfoPtr;
    if (!checkSystemInfo(systemInfo)) return false;

    systemInfo->manufacturer = manufacturerName;
    systemInfo->modelNumber = modelNumber;
    return true;
}

bool setOptionalSystemInfoStrings(s_SystemInfo **systemInfoPtr, char *serialNumber, char *firmware, char *hardware, char *software)
{
    s_SystemInfo* systemInfo = *systemInfoPtr;
    if (!checkSystemInfo(systemInfo)) return false;
    if (serialNumber != NULL)
    {
        systemInfo->serialNo = serialNumber;
        systemInfo->flagsSerialNo = SYSINFO_FLAGS_SERIAL_NO;
    }
    if (firmware != NULL)
    {
        systemInfo->firmware = firmware;
        systemInfo->flagsFirmware = SYSINFO_FLAGS_FIRMWARE;
    }
    if (hardware != NULL)
    {
        systemInfo->hardware = hardware;
        systemInfo->flagsHardware = SYSINFO_FLAGS_HARDWARE;
    }
    if (software != NULL)
    {
        systemInfo->software = software;
        systemInfo->flagsSoftware = SYSINFO_FLAGS_SOFTWARE;
    }
    return true;
}

/**
 *  The set value 'isRegulated' is only used if 'hasRegulationStatus' is true.
 */
bool setRegulationStatus(s_SystemInfo **systemInfoPtr, bool hasRegulationStatus, bool isRegulated)
{
    s_SystemInfo* systemInfo = *systemInfoPtr;
    if (!checkSystemInfo(systemInfo)) return false;
    
    if (hasRegulationStatus)
    {
        systemInfo->flagsRegStatus = SYSINFO_FLAGS_REG_STATUS;
        systemInfo->regulationStatus = isRegulated ? 0 : 0x8000;
    }
    return true;
}

/**
 * 16-digit HEX string. This method will put the systme id in little endian order as bytes.
 */
bool setSystemIdentifier(s_SystemInfo **systemInfoPtr, char *systemId)
{
    s_SystemInfo* systemInfo = *systemInfoPtr;
    if (!checkSystemInfo(systemInfo)) return false;
    if (!hexToLittleEndianByte(systemId, systemInfo->systemId))
    {
        NRF_LOG_DEBUG("System Id was not properly encoded. Value will be all 0's.");
        return false;
    }
    return true;
}

bool setSystemIdentifierByte(s_SystemInfo **systemInfoPtr, unsigned char *systemId)
{
    s_SystemInfo* systemInfo = *systemInfoPtr;
    if (systemId == NULL) return false;
    memcpy(systemInfo->systemId, systemId, 8);
    return true;
}

#if (USES_SYSTEM_AVAS == 1)
bool initializeSystemInfoAvas(s_SystemInfo **systemInfoPtr, unsigned short numberOfAvas)
{
    s_SystemInfo* systemInfo = *systemInfoPtr;
    if (!checkSystemInfo(systemInfo)) return false;

    systemInfo->numberOfAvas = numberOfAvas;
    systemInfo->currentAvaCount = 0;
    if (numberOfAvas > 0)
    {
        if (!allocateAvas(&systemInfo->avas, numberOfAvas))
        {
            systemInfo->numberOfAvas = 0;
        }
    }
    return true;
}

bool addSystemInfoAvas(s_SystemInfo **systemInfoPtr, s_Avas *avaIn)
{
    s_SystemInfo* systemInfo = *systemInfoPtr;
    if (!checkSystemInfo(systemInfo)) return false;
    if (systemInfo->numberOfAvas == 0)
    {
        NRF_LOG_DEBUG("SystemInfo AVAs not initialized or Ava initialization failed.");
        return false;
    }
    if (systemInfo->currentAvaCount >= systemInfo->numberOfAvas)
    {
        NRF_LOG_DEBUG("Number of System Info Avas have been exceeded those set during initialization. Cannot add any more.");
        return false;
    }
    if (addAva(&systemInfo->avas[systemInfo->currentAvaCount], avaIn))
    {
        systemInfo->currentAvaCount++;
        systemInfo->flagsAvas = SYSINFO_FLAGS_AVAS;
        return true;
    }
    return false;
}
#endif

static int loadString(char *str, unsigned char *systemInfoBuf, int index)
{
    if (str != NULL)
    {
        unsigned char len = (unsigned char)strlen(str);
        systemInfoBuf[index++] = (unsigned char)(len & 0xFF);
        memcpy(&systemInfoBuf[index], str, len);
        return (index + len);
    }
    return index;
}

bool createSystemInfoData(s_SystemInfoData** systemInfoDataPtr, s_SystemInfo *systemInfo)
{
    s_SystemInfoData* systemInfoData = *systemInfoDataPtr;
    if (!checkSystemInfo(systemInfo)) return false;
    
    int i;
    unsigned short systemInfoLength = 14;   // command 2, flags 2, length 2, system id(8)
    unsigned short index = 0;
    if (systemInfoData != NULL)
    {
        cleanUpSystemInfoData(systemInfoDataPtr);
    }
    systemInfoData = (s_SystemInfoData *)calloc(1, sizeof(s_SystemInfoData));
    if (systemInfoData == NULL)
    {
        NRF_LOG_DEBUG("Could not allocate memory for the System Info Index struct.");
        return false;
    }
    unsigned short flags = (unsigned short)(systemInfo->flagsAvas | systemInfo->flagsFirmware | systemInfo->flagsHardware 
                            | systemInfo->flagsSoftware | systemInfo->flagsSerialNo | systemInfo->flagsRegStatus);
    systemInfoLength = systemInfoLength + 1 + systemInfo->currentSpecializationCount * 4 +
        ((systemInfo->manufacturer != NULL) ? (1 + (unsigned short)strlen(systemInfo->manufacturer)) : 0) +
        ((systemInfo->modelNumber != NULL) ? (1 + (unsigned short)strlen(systemInfo->modelNumber)) : 0) +
    (((flags & SYSINFO_FLAGS_FIRMWARE) == SYSINFO_FLAGS_FIRMWARE) ? (1 + (unsigned short)strlen(systemInfo->firmware)) : 0) +
    (((flags & SYSINFO_FLAGS_HARDWARE) == SYSINFO_FLAGS_HARDWARE) ? (1 + (unsigned short)strlen(systemInfo->hardware)) : 0) +
    (((flags & SYSINFO_FLAGS_SOFTWARE) == SYSINFO_FLAGS_SOFTWARE) ? (1 + (unsigned short)strlen(systemInfo->software)) : 0) +
    (((flags & SYSINFO_FLAGS_SERIAL_NO) == SYSINFO_FLAGS_SERIAL_NO) ? (1 + (unsigned short)strlen(systemInfo->serialNo)) : 0) +
    (((flags & SYSINFO_FLAGS_REG_STATUS) == SYSINFO_FLAGS_REG_STATUS) ? 2 : 0);
    #if (USES_SYSTEM_AVAS == 1)
    if (systemInfo->flagsAvas != 0)
    {
        unsigned short j;
        systemInfoLength++; // for the number of AVAs entry
        for (j = 0; j < systemInfo->currentAvaCount; j++)
        {
            systemInfoLength = systemInfoLength + systemInfo->avas[j]->length + 6;  // 4-byte id and 2-byte length + 1 byte # of AVAs
        }
    }
    #endif
    systemInfoData->systemInfoBuf = (unsigned char *)calloc(1, systemInfoLength * sizeof(unsigned char));
    if (systemInfoData->systemInfoBuf == NULL)
    {
        NRF_LOG_DEBUG("Could not allocate memory for the System Info data buffer.");
        free(systemInfoData);
        return false;
    }
    // Command
    index = twoByteEncode(systemInfoData->systemInfoBuf, index, COMMAND_GET_SYS_INFO);
    // flags
    index = twoByteEncode(systemInfoData->systemInfoBuf, index, flags);
    // PDU length
    systemInfoData->dataLength = (unsigned short)systemInfoLength;
    index = twoByteEncode(systemInfoData->systemInfoBuf, index, (systemInfoLength - 6));
    // System Id
    memcpy(&systemInfoData->systemInfoBuf[index], systemInfo->systemId, 8);
    index = index + 8;
    // Specializations
    systemInfoData->systemInfoBuf[index++] = (unsigned char)(systemInfo->currentSpecializationCount & 0xFF);
    for (i = 0; i < systemInfo->currentSpecializationCount; i++)
    {
        index = twoByteEncode(systemInfoData->systemInfoBuf, index, systemInfo->specializations[i].specialization);
        index = twoByteEncode(systemInfoData->systemInfoBuf, index, systemInfo->specializations[i].version);
    }
    // Manufacturer name
    index = loadString(systemInfo->manufacturer, systemInfoData->systemInfoBuf, index);
    // Model Number
    index = loadString(systemInfo->modelNumber, systemInfoData->systemInfoBuf, index);
    // Reg status
    if (systemInfo->flagsRegStatus != 0)
    {
        systemInfoData->systemInfoBuf[index++] = (systemInfo->regulationStatus & 0xFF);
        systemInfoData->systemInfoBuf[index++] = ((systemInfo->regulationStatus >> 8) & 0xFF);
    }
    // Serial Number
    if (systemInfo->flagsSerialNo != 0)
    {
        index = loadString(systemInfo->serialNo, systemInfoData->systemInfoBuf, index);
    }
    // Firmware
    if (systemInfo->flagsFirmware != 0)
    {
        index = loadString(systemInfo->firmware, systemInfoData->systemInfoBuf, index);
    }
    // Software
    if (systemInfo->flagsSoftware != 0)
    {
        index = loadString(systemInfo->software, systemInfoData->systemInfoBuf, index);
    }
    // Hardware
    if (systemInfo->flagsHardware != 0)
    {
        index = loadString(systemInfo->hardware, systemInfoData->systemInfoBuf, index);
    }
#if (USES_SYSTEM_AVAS == 1)
    if (systemInfo->flagsAvas != 0)
    {
        int k;
        systemInfoData->systemInfoBuf[index++] = (unsigned char)systemInfo->currentAvaCount;
        NRF_LOG_DEBUG("System Info has %u AVAs\n", systemInfo->currentAvaCount);
        for (k = 0; k < systemInfo->currentAvaCount; k++)
        {
            index = fourByteEncode(systemInfoData->systemInfoBuf, index, systemInfo->avas[k]->attrId);
            systemInfoData->systemInfoBuf[index++] = (unsigned char)(systemInfo->avas[k]->length & 0xFF);
            systemInfoData->systemInfoBuf[index++] = (unsigned char)((systemInfo->avas[k]->length >> 8) & 0xFF);
            memcpy(&systemInfoData->systemInfoBuf[index], systemInfo->avas[k]->value, (size_t)(systemInfo->avas[k]->length));
            index = index + systemInfo->avas[k]->length;
        }
    }
#endif
    *systemInfoDataPtr = systemInfoData;
    return true;
}

void cleanUpGhsMsmt(s_GhsMsmt** ghsMsmtPtr)
{
    int i;
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    #if (USES_MSMT_AVAS == 1)
    cleanUpAvas(ghsMsmt->avas, ghsMsmt->numberOfAvas);
    #endif
    ghsMsmt->avas = NULL;
    if (ghsMsmt->srcHandleRefs != NULL)
    {
        free(ghsMsmt->srcHandleRefs);
    }
    if (ghsMsmt->suppTypes != NULL)
    {
        free(ghsMsmt->suppTypes);
    }
    if (ghsMsmt->duration != NULL)
    {
        free(ghsMsmt->duration);
    }
    if (ghsMsmt->compoundNumeric != NULL)
    {
        for (i = 0; i < ghsMsmt->compoundNumeric->numberOfComponents; i++)
        {
            free(ghsMsmt->compoundNumeric->value[i]);
        }
        free(ghsMsmt->compoundNumeric->value);
        free(ghsMsmt->compoundNumeric);
    }
    if (ghsMsmt->rtsa != NULL)
    {
        if (ghsMsmt->rtsa->samples != NULL)
        {
            free(ghsMsmt->rtsa->samples);
        }
        free(ghsMsmt->rtsa);
    }
    if (ghsMsmt->simpleNumeric != NULL)
    {
        free(ghsMsmt->simpleNumeric);
    }
    if (ghsMsmt->codedEnum != NULL)
    {
        free(ghsMsmt->codedEnum);
    }
    if (ghsMsmt->bits16Enum != NULL)
    {
        free(ghsMsmt->bits16Enum);
    }
    if (ghsMsmt->bits32Enum != NULL)
    {
        free(ghsMsmt->bits32Enum);
    }
    free(ghsMsmt);
    ghsMsmt = NULL;
    *ghsMsmtPtr = NULL;
}

static void cleanUpHeader(s_MsmtHeader **headerPtr)
{
    s_MsmtHeader* header = *headerPtr;
    if (header != NULL)
    {
        #if (USES_HEADER_AVAS == 1)
        cleanUpAvas(header->avas, header->numberOfAvas);
        header->avas = NULL;
        #endif
        if (header->duration != NULL)
        {
            free(header->duration);
            header->duration = NULL;
        }
        if (header->suppTypes != NULL)
        {
            free(header->suppTypes);
            header->suppTypes = NULL;
        }
        if (header->srcHandleRefs != NULL)
        {
            free(header->srcHandleRefs);
            header->srcHandleRefs = NULL;
        }
        free(header);
        *headerPtr = NULL;
    }
}

void cleanUpMsmtGroup(s_MsmtGroup** msmtGroupPtr)
{
    s_MsmtGroup* msmtGroup = *msmtGroupPtr;
    if (msmtGroup != NULL)
    {
        int i;
        for (i = 0; i < msmtGroup->header->numberOfMsmts; i++)
        {
            if (msmtGroup->ghsMsmts[i] != NULL)
            {
                cleanUpGhsMsmt(&msmtGroup->ghsMsmts[i]);
                msmtGroup->ghsMsmts[i] = NULL;
            }
        }
        if (msmtGroup->ghsMsmts != NULL)
        {
            free(msmtGroup->ghsMsmts);
            msmtGroup->ghsMsmts = NULL;
        }
        cleanUpHeader(&msmtGroup->header);
        free(msmtGroup);
        msmtGroup = NULL;
        *msmtGroupPtr = NULL;
    }
}

void cleanUpMsmtGroupData(s_MsmtGroupData** msmtGroupDataPtr)
{
    s_MsmtGroupData* msmtGroupData = *msmtGroupDataPtr;
    if (msmtGroupData != NULL)
    {
        if (msmtGroupData->ghsMsmtIndex != NULL)
        {
            int i;
            for (i = 0; i < msmtGroupData->currentGhsMsmtCount; i++)
            {
                if (msmtGroupData->ghsMsmtIndex[i] != NULL)
                {
                    free(msmtGroupData->ghsMsmtIndex[i]);
                }
            }
            free(msmtGroupData->ghsMsmtIndex);
        }
        if (msmtGroupData->data != NULL)
        {
            free(msmtGroupData->data);
        }
        free(msmtGroupData);
        *msmtGroupDataPtr = NULL;
    }
}

// ======================= Measurements
bool createMsmtGroup(s_MsmtGroup **msmtGroupPtr, bool hasTimeStamp, unsigned char numberOfGhsMsmts, unsigned char groupId)
{
    if (numberOfGhsMsmts == 0)
    {
        return false;
    }
    s_MsmtGroup* msmtGroup = *msmtGroupPtr;
    if (msmtGroup != NULL)
    {
        cleanUpMsmtGroup(msmtGroupPtr);
    }
    msmtGroup = (s_MsmtGroup *)calloc(1, sizeof(s_MsmtGroup));
    if (msmtGroup == NULL)
    {
        NRF_LOG_DEBUG("Could not allocate memory for the s_MsmtGroup.");
        return false;
    }
    msmtGroup->ghsMsmts = (s_GhsMsmt **)calloc(1, numberOfGhsMsmts * sizeof(s_GhsMsmt *));
    if (msmtGroup->ghsMsmts == NULL)
    {
        NRF_LOG_DEBUG("Could not allocate memory for the %d Msmt group s_GhsMsmt pointers.", numberOfGhsMsmts);
        free (msmtGroup);
        return false;
    }
    msmtGroup->header = (s_MsmtHeader *)calloc(1, sizeof(s_MsmtHeader));
    if (msmtGroup->header == NULL)
    {
        NRF_LOG_DEBUG("Could not allocate memory for the s_MsmtHeader struct.");
        free (msmtGroup->ghsMsmts);
        free (msmtGroup);
        return false;
    }
    msmtGroup->header->groupId = groupId;
    msmtGroup->header->numberOfMsmts = numberOfGhsMsmts;
    msmtGroup->header->flagsTimeStamp = hasTimeStamp ? HEADER_FLAGS_TIMESTAMP : 0;
    *msmtGroupPtr = msmtGroup;
    return true;
}

bool setHeaderOptions(s_MsmtGroup** msmtGroupPtr, bool areSettings, bool hasPersonId, unsigned short personId)
{
    s_MsmtGroup* msmtGroup = *msmtGroupPtr;
    if (!checkMsmtGroup(msmtGroup)) return false;

    msmtGroup->header->flagsPersonId = hasPersonId ? HEADER_FLAGS_PERSONID : 0;
    if (hasPersonId)
    {
        msmtGroup->header->personId = personId;
    }
    msmtGroup->header->flagsSetting = areSettings ? HEADER_FLAGS_SETTING : 0;
    return true;
}

bool setHeaderSupplementalTypes(s_MsmtGroup **msmtGroupPtr, unsigned long *supplementalTypes, unsigned short numberOfSupplementalTypes )
{
    s_MsmtGroup* msmtGroup = *msmtGroupPtr;
    if (!checkMsmtGroup(msmtGroup)) return false;

    if (numberOfSupplementalTypes > 0)
    {
        int i;
        if (supplementalTypes == NULL)
        {
            NRF_LOG_DEBUG("Supplemental types are NULL.");
            return false;
        }
        if (msmtGroup->header->suppTypes != NULL)
        {
            free (msmtGroup->header->suppTypes);
            msmtGroup->header->flagsSuppTypes = 0;
            msmtGroup->header->numberOfSuppTypes = 0;
        }
        msmtGroup->header->suppTypes = (unsigned long *)calloc(1, numberOfSupplementalTypes * sizeof(unsigned long));
        if (msmtGroup->header->suppTypes == NULL)
        {
            NRF_LOG_DEBUG("Could not allocate memory for the Header supplemental types.");
            return false;
        }
        msmtGroup->header->numberOfSuppTypes = numberOfSupplementalTypes;
        msmtGroup->header->flagsSuppTypes = HEADER_FLAGS_SUPP_TYPES;
        for (i = 0; i < numberOfSupplementalTypes; i++)
        {
            msmtGroup->header->suppTypes[i] = supplementalTypes[i];
        }
    }
    return true;
}

bool setHeaderSrcHandleRefs(s_MsmtGroup **msmtGroupPtr, unsigned short *srcHandleRefs, unsigned short numberOfSrcHandleRefs )
{
    s_MsmtGroup* msmtGroup = *msmtGroupPtr;
    if (!checkMsmtGroup(msmtGroup)) return false;

    if (numberOfSrcHandleRefs > 0)
    {
        if (srcHandleRefs == NULL)
        {
            NRF_LOG_DEBUG("Source Handle references are NULL.");
            return false;
        }
        if (msmtGroup->header->srcHandleRefs != NULL)
        {
            free (msmtGroup->header->srcHandleRefs);
            msmtGroup->header->numberOfSrcHandleRefs = 0;
            msmtGroup->header->flagsSrcHndlRef = 0;
        }
        msmtGroup->header->srcHandleRefs = (unsigned short *)calloc(1, numberOfSrcHandleRefs * sizeof(unsigned short));
        if (msmtGroup->header->srcHandleRefs == NULL)
        {
            NRF_LOG_DEBUG("Could not allocate memory for the Header src handle refs.");
            return false;
        }
        int i;
        for (i = 0; i < numberOfSrcHandleRefs; i++)
        {
            msmtGroup->header->srcHandleRefs[i] = srcHandleRefs[i];
        }
        msmtGroup->header->numberOfSrcHandleRefs = numberOfSrcHandleRefs;
        msmtGroup->header->flagsSrcHndlRef = MSMT_FLAGS_SRC_HNDL_REF;
    }
    return true;
}

bool setHeaderDuration(s_MsmtGroup **msmtGroupPtr, s_MderFloat *duration )
{
    s_MsmtGroup* msmtGroup = *msmtGroupPtr;
    if (!checkMsmtGroup(msmtGroup)) return false;

    if (duration == NULL)
    {
        NRF_LOG_DEBUG("Duration is NULL.");
        return false;
    }
    if (msmtGroup->header->duration != NULL)
    {
        free(msmtGroup->header->duration);
        msmtGroup->header->flagsDuration = 0;
    }
    
    msmtGroup->header->duration = (s_MderFloat *)calloc(1, sizeof(s_MderFloat));
    if (msmtGroup->header->duration == NULL)
    {
        NRF_LOG_DEBUG("Could not allocate memory for the Header duration.");
        return false;
    }
    memcpy(msmtGroup->header->duration, duration, sizeof(s_MderFloat));
    msmtGroup->header->flagsDuration = HEADER_FLAGS_DURATION;
    return true;
}

#if (USES_HEADER_AVAS == 1)
bool initializeHeaderAvas(s_MsmtGroup **msmtGroupPtr, unsigned short numberOfAvas)
{
    s_MsmtGroup* msmtGroup = *msmtGroupPtr;
    if (!checkMsmtGroup(msmtGroup)) return false;

    msmtGroup->header->numberOfAvas = numberOfAvas;
    if (numberOfAvas > 0)
    {
        if (!allocateAvas(&msmtGroup->header->avas, numberOfAvas))
        {
            msmtGroup->header->numberOfAvas = 0;
        }
    }
    return true;
}

bool addHeaderAva(s_MsmtGroup **msmtGroupPtr, s_Avas *avaIn)
{
    s_MsmtGroup* msmtGroup = *msmtGroupPtr;
    if (!checkMsmtGroup(msmtGroup)) return false;

    if (msmtGroup->header->numberOfAvas == 0)
    {
        NRF_LOG_DEBUG("Header AVAs not initialized or Ava initialization failed.");
        return false;
    }
    if (msmtGroup->header->currentAvaCount >= msmtGroup->header->numberOfAvas)
    {
        NRF_LOG_DEBUG("Number of Header Avas have been exceeded. Cannot add any more.");
        return false;
    }
    if (addAva(&msmtGroup->header->avas[msmtGroup->header->currentAvaCount], avaIn))
    {
        msmtGroup->header->currentAvaCount++;
        return true;
    }
    return false;
}
#endif

short addGhsMsmtToGroup(s_GhsMsmt *ghsMsmt, s_MsmtGroup **msmtGroupPtr)
{
    s_MsmtGroup* msmtGroup = *msmtGroupPtr;
    if (!checkMsmtGroup(msmtGroup)) return -1;
    if (!checkGhsMsmt(ghsMsmt)) return -1;
    if (msmtGroup->header->currentMsmtCount >= msmtGroup->header->numberOfMsmts)
    {
        NRF_LOG_DEBUG("Number of GhsMsmts %d exceeds numberOfGhsMsmts %d.", (msmtGroup->header->currentMsmtCount + 1), msmtGroup->header->numberOfMsmts);
        return -1;
    }
    msmtGroup->ghsMsmts[msmtGroup->header->currentMsmtCount++] = ghsMsmt;
    return (msmtGroup->header->currentMsmtCount - 1);
}

#if (USES_NUMERIC == 1)
bool createNumericMsmt(s_GhsMsmt **ghsMsmtPtr, unsigned long type, bool isSfloat, unsigned short units)
{
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (ghsMsmt != NULL && ghsMsmt->flagsMsmtType == MSMT_FLAGS_SIMPLE_NUMERIC)
    {
        cleanUpGhsMsmt(&ghsMsmt);
    }
    ghsMsmt = (s_GhsMsmt *)calloc(1, sizeof(s_GhsMsmt));
    if (ghsMsmt == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for Numeric Ghs Measurement.");
        return false;
    }
    ghsMsmt->simpleNumeric = (s_SimpNumeric *) calloc(1, sizeof(s_SimpNumeric));
    if (ghsMsmt->simpleNumeric == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for Simple Numeric entry.");
        free (ghsMsmt);
        return false;
    }
    ghsMsmt->type = type;
    ghsMsmt->flagsMsmtType = MSMT_FLAGS_SIMPLE_NUMERIC;
    ghsMsmt->flagsSfloat = isSfloat ? MSMT_FLAGS_SFLOAT_VAL : 0;
    ghsMsmt->simpleNumeric->units = units;
    *ghsMsmtPtr = ghsMsmt;
    return true;
}

bool updateNumericMsmt(s_GhsMsmt **ghsMsmtPtr, s_MderFloat *value, unsigned short msmt_id)
{
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (!checkGhsMsmt(ghsMsmt)) return false;
    
    if (value == NULL)
    {
        NRF_LOG_DEBUG("Input value is NULL. Skipping");
        return false;
    }
    if (ghsMsmt->flagsMsmtType != MSMT_FLAGS_SIMPLE_NUMERIC)
    {
        NRF_LOG_DEBUG("Measurement is not a numeric. Skipping");
        return false;
    }
    if ((ghsMsmt->flagsSfloat == MSMT_FLAGS_SFLOAT_VAL && value->mderFloatType != MDER_SFLOAT) ||
        (ghsMsmt->flagsSfloat == 0 && value->mderFloatType != MDER_FLOAT))
    {
        NRF_LOG_DEBUG("Input Measurement float type is not that of the original measurement. Skipping");
        return false;
    }
    ghsMsmt->simpleNumeric->value = *copyMderFloat(&ghsMsmt->simpleNumeric->value, value);
    ghsMsmt->id = msmt_id;
    return true;
}

bool updateDataNumeric(s_MsmtGroupData** msmtGroupDataPtr, unsigned short msmtIndex, s_MderFloat* value, unsigned short msmt_id)
{
    s_MsmtGroupData* msmtGroupData = *msmtGroupDataPtr;
    if (!checkMsmtGroupData(msmtGroupData)) return false;

    if (value == NULL)
    {
        NRF_LOG_DEBUG("Input value is NULL. Skipping");
        return false;
    }
    s_GhsMsmtIndex* ghsMsmtIndex = msmtGroupData->ghsMsmtIndex[msmtIndex];
    if (ghsMsmtIndex->metricType != MSMT_FLAGS_SIMPLE_NUMERIC)
    {
        NRF_LOG_DEBUG("Measurement is not a numeric. Skipping");
        return false;
    }
    msmtGroupData->data[ghsMsmtIndex->id_index] = (msmt_id & 0xFF);
    msmtGroupData->data[ghsMsmtIndex->id_index + 1] = ((msmt_id >> 8) & 0xFF);
    if (ghsMsmtIndex->isSfloat)
    {
        unsigned short ieeeFloat;
        createIeeeSFloatFromMderFloat(value, &ieeeFloat);
        msmtGroupData->data[ghsMsmtIndex->value_index] = (ieeeFloat & 0xFF);
        msmtGroupData->data[ghsMsmtIndex->value_index + 1] = ((ieeeFloat >> 8) & 0xFF);
    }
    else
    {
        unsigned long ieeeFloat;
        createIeeeFloatFromMderFloat(value, &ieeeFloat);
        fourByteEncode(msmtGroupData->data, ghsMsmtIndex->value_index, ieeeFloat);
    }
    return true;
}
#endif
#if (USES_COMPOUND == 1)
bool createCompoundNumericMsmt(s_GhsMsmt** ghsMsmtPtr, unsigned long type, bool isSfloat, unsigned short units, unsigned short numberOfComponents, s_Compound *compounds)
{
    int i;
    if (compounds == NULL || numberOfComponents == 0)
    {
        NRF_LOG_DEBUG("Compounds entry was NULL or numberOfComponents 0.");
        return false;
    }
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (ghsMsmt != NULL && ghsMsmt->flagsMsmtType == MSMT_FLAGS_COMPOUND_NUMERIC)
    {
        cleanUpGhsMsmt(&ghsMsmt);
    }
    ghsMsmt = (s_GhsMsmt *)calloc(1, sizeof(s_GhsMsmt));
    if (ghsMsmt == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for Numeric Ghs Measurement.");
        return false;
    }
    ghsMsmt->compoundNumeric = (s_CompoundNumeric *) calloc(1, sizeof(s_CompoundNumeric));
    if (ghsMsmt->compoundNumeric == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for Compound Numeric Entry.");
        free (ghsMsmt);
        return false;
    }
    ghsMsmt->type = type;
    ghsMsmt->flagsMsmtType = MSMT_FLAGS_COMPOUND_NUMERIC;
    ghsMsmt->flagsSfloat = isSfloat ? MSMT_FLAGS_SFLOAT_VAL : 0;
    ghsMsmt->compoundNumeric->units = units;
    ghsMsmt->compoundNumeric->numberOfComponents = numberOfComponents;
    ghsMsmt->compoundNumeric->value = (s_Compound **)calloc(1, numberOfComponents * sizeof(s_Compound *));
    if (ghsMsmt->compoundNumeric->value == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for Compound Numeric value pointers.");
        free (ghsMsmt->compoundNumeric);
        free (ghsMsmt);
        return false;
    }
    for (i = 0; i < numberOfComponents; i++)
    {
        ghsMsmt->compoundNumeric->value[i] = (s_Compound *)calloc(1, sizeof(s_Compound));
        if (ghsMsmt->compoundNumeric->value[i] == NULL)
        {
            NRF_LOG_DEBUG("Unable to allocate memory for Compound Numeric value %d.", i);
            while (i > 0)
            {
                free (ghsMsmt->compoundNumeric->value[i - 1]);
            }
            free (ghsMsmt->compoundNumeric);
            free (ghsMsmt);
            return false;
        }
        ghsMsmt->compoundNumeric->value[i]->subType = compounds[i].subType;
        ghsMsmt->compoundNumeric->value[i]->subValue = *copyMderFloat(&ghsMsmt->compoundNumeric->value[i]->subValue, &compounds[i].subValue);
    }
    *ghsMsmtPtr = ghsMsmt;
    return true;
}

bool updateCompoundMsmt(s_GhsMsmt **ghsMsmtPtr, s_MderFloat *value, unsigned short msmt_id)
{
    int i;
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (!checkGhsMsmt(ghsMsmt)) return false;
    
    if (value == NULL)
    {
        NRF_LOG_DEBUG("Input value is NULL. Skipping");
        return false;
    }
    if (ghsMsmt->flagsMsmtType != MSMT_FLAGS_COMPOUND_NUMERIC)
    {
        NRF_LOG_DEBUG("Measurement is not a compound numeric. Skipping");
        return false;
    }
    if ((ghsMsmt->flagsSfloat == MSMT_FLAGS_SFLOAT_VAL && value[0].mderFloatType != MDER_SFLOAT) ||
        (ghsMsmt->flagsSfloat == 0 && value[0].mderFloatType != MDER_FLOAT))
    {
        NRF_LOG_DEBUG("Input Measurement float type is not that of the original measurement. Skipping");
        return false;
    }
    for (i = 0; i < ghsMsmt->compoundNumeric->numberOfComponents; i++)
    {
        ghsMsmt->compoundNumeric->value[i]->subValue = *copyMderFloat(&ghsMsmt->compoundNumeric->value[i]->subValue, &value[i]);
    }
    ghsMsmt->id = msmt_id;
    return true;
}

bool updateDataCompound(s_MsmtGroupData** msmtGroupDataPtr, unsigned short msmtIndex, s_MderFloat* value, unsigned short msmt_id)
{
    s_MsmtGroupData* msmtGroupData = *msmtGroupDataPtr;
    if (!checkMsmtGroupData(msmtGroupData)) return false;

    if (value == NULL)
    {
        NRF_LOG_DEBUG("Input value is NULL. Skipping");
        return false;
    }
    s_GhsMsmtIndex* ghsMsmtIndex = msmtGroupData->ghsMsmtIndex[msmtIndex];
    if (ghsMsmtIndex->metricType != MSMT_FLAGS_COMPOUND_NUMERIC)
    {
        NRF_LOG_DEBUG("Measurement is not a compound. Skipping");
        return false;
    }
    unsigned short i;
    msmtGroupData->data[ghsMsmtIndex->id_index] = (msmt_id & 0xFF);
    msmtGroupData->data[ghsMsmtIndex->id_index + 1] = ((msmt_id >> 8) & 0xFF);
    unsigned short index = ghsMsmtIndex->value_index;
    bool cont = (((msmtGroupData->data[2] & 0xFF) + ((msmtGroupData->data[2] << 8) & 0xFF) & HEADER_FLAGS_CONT_FOLLOWS) == HEADER_FLAGS_CONT_FOLLOWS);
    for (i = 0; i < ghsMsmtIndex->numberOfValues; i++)
    {
        if (ghsMsmtIndex->isSfloat)
        {
            unsigned short ieeeFloat;
            createIeeeSFloatFromMderFloat(&value[i], &ieeeFloat);
            msmtGroupData->data[index++] = (ieeeFloat & 0xFF);
            msmtGroupData->data[index++] = ((ieeeFloat >> 8) & 0xFF);
        }
        else
        {
            unsigned long ieeeFloat;
            createIeeeFloatFromMderFloat(&value[i], &ieeeFloat);
            index = fourByteEncode(msmtGroupData->data, index, ieeeFloat);
        }
        if (!cont) index = index + 4;
    }
    return true;
}
#endif
#if (USES_CODED == 1)
bool createCodedMsmt(s_GhsMsmt** ghsMsmtPtr, unsigned long type)
{
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (ghsMsmt != NULL && ghsMsmt->flagsMsmtType == MSMT_FLAGS_CODED_ENUM)
    {
        cleanUpGhsMsmt(&ghsMsmt);
    }
    ghsMsmt = (s_GhsMsmt *)calloc(1, sizeof(s_GhsMsmt));
    if (ghsMsmt == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for Coded Ghs Measurement.");
        return false;
    }
    ghsMsmt->codedEnum = (s_CodedEnum *) calloc(1, sizeof(s_CodedEnum));
    if (ghsMsmt->codedEnum == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for Coded Enum Entry.");
        free (ghsMsmt);
        return false;
    }
    ghsMsmt->type = type;
    ghsMsmt->flagsMsmtType = MSMT_FLAGS_CODED_ENUM;
    *ghsMsmtPtr = ghsMsmt;
    return true;
}

bool updateCodedEnum(s_GhsMsmt **ghsMsmtPtr, unsigned long code, unsigned short msmt_id)
{
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (!checkGhsMsmt(ghsMsmt)) return false;
    
    if (ghsMsmt->flagsMsmtType != MSMT_FLAGS_CODED_ENUM)
    {
        NRF_LOG_DEBUG("Measurement is not a Coded Enum. Skipping");
        return false;
    }
    ghsMsmt->codedEnum->code = code;
    ghsMsmt->id = msmt_id;
    return true;
}

bool updateDataCoded(s_MsmtGroupData** msmtGroupDataPtr, unsigned short msmtIndex, unsigned long code, unsigned short msmt_id)
{
    s_MsmtGroupData* msmtGroupData = *msmtGroupDataPtr;
    if (!checkMsmtGroupData(msmtGroupData)) return false;

    s_GhsMsmtIndex* ghsMsmtIndex = msmtGroupData->ghsMsmtIndex[msmtIndex];
    if (ghsMsmtIndex->metricType != MSMT_FLAGS_CODED_ENUM)
    {
        NRF_LOG_DEBUG("Measurement is not a coded enum. Skipping");
        return false;
    }
    msmtGroupData->data[ghsMsmtIndex->id_index] = (msmt_id & 0xFF);
    msmtGroupData->data[ghsMsmtIndex->id_index + 1] = ((msmt_id >> 8) & 0xFF);
    fourByteEncode(msmtGroupData->data, ghsMsmtIndex->value_index, code);
    return true;
}
#endif

#if (USES_BITS16 == 1)
bool createBit16EnumMsmt(s_GhsMsmt** ghsMsmtPtr, unsigned long type, unsigned short state, unsigned short support)
{
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (ghsMsmt != NULL && ghsMsmt->flagsMsmtType == MSMT_FLAGS_16BITS_ENUM)
    {
        cleanUpGhsMsmt(&ghsMsmt);
    }
    ghsMsmt = (s_GhsMsmt*)calloc(1, sizeof(s_GhsMsmt));
    if (ghsMsmt == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for BITS 16 Ghs Measurement.");
        return false;
    }
    ghsMsmt->type = type;
    ghsMsmt->bits16Enum = (s_Bits16Enum*)calloc(1, sizeof(s_Bits16Enum));
    if (ghsMsmt->bits16Enum == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for BITs 16 Enum Entry.");
        free(ghsMsmt);
        return false;
    }
    ghsMsmt->bits16Enum->stateEvent = state;
    ghsMsmt->bits16Enum->supportEvent = support;
    ghsMsmt->flagsMsmtType = MSMT_FLAGS_16BITS_ENUM;
    *ghsMsmtPtr = ghsMsmt;
    return true;
}

bool updateBits16Enum(s_GhsMsmt** ghsMsmtPtr, unsigned short bits16, unsigned short msmt_id)
{
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (!checkGhsMsmt(ghsMsmt)) return false;

    if (ghsMsmt->flagsMsmtType != MSMT_FLAGS_16BITS_ENUM)
    {
        NRF_LOG_DEBUG("Measurement is not a BITs 16 Enum. Skipping");
        return false;
    }
    ghsMsmt->bits16Enum->bitsEnum = bits16;
    ghsMsmt->id = msmt_id;
    return true;
}

bool updateDataBits16(s_MsmtGroupData** msmtGroupDataPtr, unsigned short msmtIndex, unsigned short bits16, unsigned short msmt_id)
{
    s_MsmtGroupData* msmtGroupData = *msmtGroupDataPtr;
    if (!checkMsmtGroupData(msmtGroupData)) return false;

    s_GhsMsmtIndex* ghsMsmtIndex = msmtGroupData->ghsMsmtIndex[msmtIndex];
    if (ghsMsmtIndex->metricType != MSMT_FLAGS_16BITS_ENUM)
    {
        NRF_LOG_DEBUG("Measurement is not a BITS 16 enum. Skipping");
        return false;
    }
    msmtGroupData->data[ghsMsmtIndex->id_index] = (msmt_id & 0xFF);
    msmtGroupData->data[ghsMsmtIndex->id_index + 1] = ((msmt_id >> 8) & 0xFF);
    twoByteEncode(msmtGroupData->data, ghsMsmtIndex->value_index, bits16);
    return true;
}
#endif
#if (USES_BITS32 == 1)
bool createBit32EnumMsmt(s_GhsMsmt** ghsMsmtPtr, unsigned long type, unsigned long state, unsigned long support)
{
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (ghsMsmt != NULL && ghsMsmt->flagsMsmtType == MSMT_FLAGS_32BITS_ENUM)
    {
        cleanUpGhsMsmt(&ghsMsmt);
    }
    ghsMsmt = (s_GhsMsmt*)calloc(1, sizeof(s_GhsMsmt));
    if (ghsMsmt == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for BITS 32 Ghs Measurement.");
        return false;
    }
    ghsMsmt->type = type;
    ghsMsmt->bits32Enum = (s_Bits32Enum*)calloc(1, sizeof(s_Bits32Enum));
    if (ghsMsmt->bits32Enum == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for BITs 32 Enum Entry.");
        free(ghsMsmt);
        return false;
    }
    ghsMsmt->bits32Enum->stateEvent = state;
    ghsMsmt->bits32Enum->supportEvent = support;
    ghsMsmt->flagsMsmtType = MSMT_FLAGS_32BITS_ENUM;
    *ghsMsmtPtr = ghsMsmt;
    return true;
}

bool updateBits32Enum(s_GhsMsmt** ghsMsmtPtr, unsigned long bits32, unsigned short msmt_id)
{
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (!checkGhsMsmt(ghsMsmt)) return false;

    if (ghsMsmt->flagsMsmtType != MSMT_FLAGS_32BITS_ENUM)
    {
        NRF_LOG_DEBUG("Measurement is not a BITs 32 Enum. Skipping");
        return false;
    }
    ghsMsmt->bits32Enum->bitsEnum = bits32;
    ghsMsmt->id = msmt_id;
    return true;
}

bool updateDataBits32(s_MsmtGroupData** msmtGroupDataPtr, unsigned short msmtIndex, unsigned long bits32, unsigned short msmt_id)
{
    s_MsmtGroupData* msmtGroupData = *msmtGroupDataPtr;
    if (!checkMsmtGroupData(msmtGroupData)) return false;

    s_GhsMsmtIndex* ghsMsmtIndex = msmtGroupData->ghsMsmtIndex[msmtIndex];
    if (ghsMsmtIndex->metricType != MSMT_FLAGS_32BITS_ENUM)
    {
        NRF_LOG_DEBUG("Measurement is not a BITS 32 enum. Skipping");
        return false;
    }
    msmtGroupData->data[ghsMsmtIndex->id_index] = (msmt_id & 0xFF);
    msmtGroupData->data[ghsMsmtIndex->id_index + 1] = ((msmt_id >> 8) & 0xFF);
    fourByteEncode(msmtGroupData->data, ghsMsmtIndex->value_index, bits32);
    return true;
}
#endif
#if (USES_RTSA == 1)
bool createRtsaMsmt(s_GhsMsmt** ghsMsmtPtr, unsigned long type, unsigned short units, s_MderFloat *period, s_MderFloat *scaleFactor, s_MderFloat *offset, 
    unsigned short numberOfSamples, unsigned char sampleSize)
{
    if (period == NULL || scaleFactor == NULL || offset == NULL)
    {
        NRF_LOG_DEBUG("One or more of the s_MderFloat Input parameters are NULL.");
        return false;
    }
    if (numberOfSamples == 0 || sampleSize == 0)
    {
        NRF_LOG_DEBUG("Either the sample size or the number of samples is 0.");
        return false;
    }
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (ghsMsmt != NULL && ghsMsmt->flagsMsmtType == MSMT_FLAGS_RTSA)
    {
        cleanUpGhsMsmt(&ghsMsmt);
    }
    ghsMsmt = (s_GhsMsmt*)calloc(1, sizeof(s_GhsMsmt));
    if (ghsMsmt == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for RTSA Ghs Measurement.");
        return false;
    }
    ghsMsmt->type = type;
    ghsMsmt->rtsa = (s_Rtsa*)calloc(1, sizeof(s_Rtsa));
    if (ghsMsmt->rtsa == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for RTSA Entry.");
        free(ghsMsmt);
        return false;
    }
    ghsMsmt->rtsa->samples = (unsigned char*)calloc(1, numberOfSamples * sampleSize);
    if (ghsMsmt->rtsa->samples == NULL)
    {
        NRF_LOG_DEBUG("Unable to allocate memory for the RTSA data sample array.");
        free(ghsMsmt);
        return false;
    }
    ghsMsmt->rtsa->offset = *copyMderFloat(&ghsMsmt->rtsa->offset, offset);
    ghsMsmt->rtsa->period = *copyMderFloat(&ghsMsmt->rtsa->period, period);
    ghsMsmt->rtsa->scaleFactor = *copyMderFloat(&ghsMsmt->rtsa->scaleFactor, scaleFactor);
    ghsMsmt->rtsa->units = units;
    ghsMsmt->rtsa->numberOfSamples = numberOfSamples;
    ghsMsmt->rtsa->sampleSize = sampleSize;
    ghsMsmt->flagsMsmtType = MSMT_FLAGS_RTSA;
    *ghsMsmtPtr = ghsMsmt;
    return true;
}

bool updateRtsa(s_GhsMsmt** ghsMsmtPtr, unsigned char *samples, unsigned short msmt_id)
{
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (!checkGhsMsmt(ghsMsmt)) return false;

    if (ghsMsmt->flagsMsmtType != MSMT_FLAGS_RTSA)
    {
        NRF_LOG_DEBUG("Measurement is not an RTSA. Skipping");
        return false;
    }
    memcpy(ghsMsmt->rtsa->samples, samples, ghsMsmt->rtsa->numberOfSamples * ghsMsmt->rtsa->sampleSize);
    ghsMsmt->id = msmt_id;
    return true;
}

bool updateDataRtsa(s_MsmtGroupData** msmtGroupDataPtr, unsigned short msmtIndex, unsigned char* samples, unsigned short dataLength, unsigned short msmt_id)
{
    s_MsmtGroupData* msmtGroupData = *msmtGroupDataPtr;
    if (!checkMsmtGroupData(msmtGroupData)) return false;

    s_GhsMsmtIndex* ghsMsmtIndex = msmtGroupData->ghsMsmtIndex[msmtIndex];
    if (ghsMsmtIndex->metricType != MSMT_FLAGS_RTSA)
    {
        NRF_LOG_DEBUG("Measurement is not an RTSA. Skipping");
        return false;
    }
    msmtGroupData->data[ghsMsmtIndex->id_index] = (msmt_id & 0xFF);
    msmtGroupData->data[ghsMsmtIndex->id_index + 1] = ((msmt_id >> 8) & 0xFF);
    memcpy(&msmtGroupData->data[ghsMsmtIndex->value_index], samples, dataLength);
    return true;
}
#endif
bool setGhsMsmtSupplementalTypes(s_GhsMsmt **ghsMsmtPtr, unsigned long *supplementalTypes, unsigned short numberOfSupplementalTypes )
{
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (!checkGhsMsmt(ghsMsmt)) return false;

    if (numberOfSupplementalTypes > 0)
    {
        if (supplementalTypes == NULL)
        {
            NRF_LOG_DEBUG("Supplemental types are NULL.");
            return false;
        }
        if (ghsMsmt->suppTypes != NULL)
        {
            free (ghsMsmt->suppTypes);
            ghsMsmt->flagsSuppTypes = 0;
            ghsMsmt->numberOfSuppTypes = 0;
        }
        ghsMsmt->suppTypes = (unsigned long *)calloc(1, numberOfSupplementalTypes * sizeof(unsigned long));
        if (ghsMsmt->suppTypes == NULL)
        {
            NRF_LOG_DEBUG("Could not allocate memory for the Ghs Msmt supplemental types.");
            return false;
        }
        ghsMsmt->numberOfSuppTypes = numberOfSupplementalTypes;
        ghsMsmt->flagsSuppTypes = MSMT_FLAGS_SUPP_TYPES;
        int i;
        for (i = 0; i < numberOfSupplementalTypes; i++)
        {
            ghsMsmt->suppTypes[i] = supplementalTypes[i];
        }
        return true;
    }
    NRF_LOG_DEBUG("Number of supplemental types was zero.");
    return false;
}

bool setGhsMsmtSrcHandleRefs(s_GhsMsmt **ghsMsmtPtr, unsigned short *srcHandleRefs, unsigned short numberOfSrcHandleRefs )
{
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (!checkGhsMsmt(ghsMsmt)) return false;

    if (numberOfSrcHandleRefs > 0)
    {
        if (srcHandleRefs == NULL)
        {
            NRF_LOG_DEBUG("Source Handle references are NULL.");
            return false;
        }
        if (ghsMsmt->srcHandleRefs != NULL)
        {
            free (ghsMsmt->srcHandleRefs);
            ghsMsmt->numberOfSrcHandleRefs = 0;
            ghsMsmt->flagsSrcHndlRef = 0;
        }
        ghsMsmt->srcHandleRefs = (unsigned short *)calloc(1, numberOfSrcHandleRefs * sizeof(unsigned short));
        if (ghsMsmt->srcHandleRefs == NULL)
        {
            NRF_LOG_DEBUG("Could not allocate memory for the Ghs Msmt src handle refs.");
            return false;
        }
        int i;
        for (i = 0; i < numberOfSrcHandleRefs; i++)
        {
            ghsMsmt->srcHandleRefs[i] = srcHandleRefs[i];
        }
        ghsMsmt->numberOfSrcHandleRefs = numberOfSrcHandleRefs;
        ghsMsmt->flagsSrcHndlRef = MSMT_FLAGS_SRC_HNDL_REF;
        return true;
    }
    NRF_LOG_DEBUG("Number of source handle references was zero.");
    return false;
}

bool setGhsMsmtDuration(s_GhsMsmt **ghsMsmtPtr, s_MderFloat *duration )
{
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (!checkGhsMsmt(ghsMsmt)) return false;

    if (duration == NULL)
    {
        NRF_LOG_DEBUG("Duration is NULL.");
        return false;
    }
    if (ghsMsmt->duration != NULL)
    {
        free(ghsMsmt->duration);
        ghsMsmt->flagsDuration = 0;
    }
    
    ghsMsmt->duration = (s_MderFloat *)calloc(1, sizeof(s_MderFloat));
    if (ghsMsmt->duration == NULL)
    {
        NRF_LOG_DEBUG("Could not allocate memory for the Ghs Msmt duration.");
        return false;
    }
    memcpy(ghsMsmt->duration, duration, sizeof(s_MderFloat));
    ghsMsmt->flagsDuration = MSMT_FLAGS_DURATION;
    return true;
}

#if (USES_MSMT_AVAS == 1)
bool initializeGhsMsmtAvas(s_GhsMsmt **ghsMsmtPtr, unsigned short numberOfAvas)
{
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (!checkGhsMsmt(ghsMsmt)) return false;

    ghsMsmt->numberOfAvas = numberOfAvas;
    if (numberOfAvas > 0)
    {
        if (!allocateAvas(&ghsMsmt->avas, numberOfAvas))
        {
            ghsMsmt->numberOfAvas = 0;
        }
    }
    return true;
}

bool addGhsMsmtAva(s_GhsMsmt **ghsMsmtPtr, s_Avas *avaIn)
{
    s_GhsMsmt* ghsMsmt = *ghsMsmtPtr;
    if (!checkGhsMsmt(ghsMsmt)) return false;

    if (ghsMsmt->numberOfAvas == 0)
    {
        NRF_LOG_DEBUG("GhsMsmt AVAs not initialized or Ava initialization failed.");
        return false;
    }
    if (ghsMsmt->currentAvaCount >= ghsMsmt->numberOfAvas)
    {
        NRF_LOG_DEBUG("Number of Ghs Msmt Avas have been exceeded. Cannot add any more.");
        return false;
    }
    if (addAva(&ghsMsmt->avas[ghsMsmt->currentAvaCount], avaIn))
    {
        ghsMsmt->currentAvaCount++;
        ghsMsmt->flagsAvas = MSMT_FLAGS_AVAS;
        return true;
    }
    return false;
}
#endif

// ===================== Measurement Data Array
static int encodeAlwaysGhsBase(int index, s_GhsMsmt* ghsMsmt, unsigned char* msmtBuf, s_GhsMsmtIndex **ghsMsmtIndex)
{
    // type
    index = fourByteEncode(msmtBuf, index, ghsMsmt->type);
    // leave spec for length
    index = index + 2;
    // flags
    int flags = (ghsMsmt->flagsMsmtType
        | ghsMsmt->flagsAvas
        | ghsMsmt->flagsSfloat
        | ghsMsmt->flagsDuration
        | ghsMsmt->flagsSrcHndlRef
        | ghsMsmt->flagsSuppTypes);
    index = twoByteEncode(msmtBuf, index, flags);
    // handle (id)
    (*ghsMsmtIndex)->id_index = index;  // Position of the msmt id
    index = twoByteEncode(msmtBuf, index, ghsMsmt->id);
    return index;
}

static int encodeOptionals(int index, s_GhsMsmt * ghsMsmt, unsigned char *msmtBuf, unsigned short *msmtLengths)
{
    int k;
    // supplemental types
    if (ghsMsmt->numberOfSuppTypes > 0)
    {
        msmtBuf[index++] = (unsigned char)(ghsMsmt->numberOfSuppTypes & 0xFF);
        for (k = 0; k < ghsMsmt->numberOfSuppTypes; k++)
        {
            index = fourByteEncode(msmtBuf, index, ghsMsmt->suppTypes[k]);
        }
        *msmtLengths = *msmtLengths + 1 + 4 * ghsMsmt->numberOfSuppTypes;
    }
    // src-handle ref
    if (ghsMsmt->flagsSrcHndlRef != 0)
    {
        msmtBuf[index++] = (unsigned char)(ghsMsmt->numberOfSrcHandleRefs & 0xFF);
        for (k = 0; k < ghsMsmt->numberOfSrcHandleRefs; k++)
        {
            index = twoByteEncode(msmtBuf, index, ghsMsmt->srcHandleRefs[k]);
        }
        *msmtLengths = *msmtLengths + 1 + 2 * ghsMsmt->numberOfSrcHandleRefs;
    }
    // duration
    if (ghsMsmt->flagsDuration != 0)
    {
        unsigned long mder;
        createIeeeFloatFromMderFloat(ghsMsmt->duration, &mder);
        index = fourByteEncode(msmtBuf, index, mder);
        *msmtLengths = *msmtLengths + 4;
    }
    #if (USES_MSMT_AVAS == 1)
    if (ghsMsmt->flagsAvas != 0)
    {
        msmtBuf[index++] = (unsigned char)ghsMsmt->currentAvaCount;
        *msmtLengths = *msmtLengths + 1;
        NRF_LOG_DEBUG("Msmt has %u AVAs\n", ghsMsmt->currentAvaCount);
        for (k = 0; k < ghsMsmt->currentAvaCount; k++)
        {
            unsigned short len = 6 + ghsMsmt->avas[k]->length;
            index = fourByteEncode(msmtBuf, index, ghsMsmt->avas[k]->attrId);
            index = twoByteEncode(msmtBuf, index, ghsMsmt->avas[k]->length);
            memcpy(&msmtBuf[index], ghsMsmt->avas[k]->value, (size_t)(ghsMsmt->avas[k]->length));
            index = index + ghsMsmt->avas[k]->length;
            *msmtLengths = *msmtLengths + len;
        }
    }
    #endif
    return index;
}

static unsigned short sizeOfGhsBase(s_GhsMsmt* ghsMsmt)
{
    unsigned short groupLength = ((ghsMsmt->flagsDuration != 0) ? 4 : 0) +
                                   ((ghsMsmt->flagsSuppTypes != 0) ? (1 + ghsMsmt->numberOfSuppTypes * 4) : 0) +
                                   ((ghsMsmt->flagsSrcHndlRef != 0) ? (1 + ghsMsmt->numberOfSrcHandleRefs * 2) : 0);
    #if (USES_MSMT_AVAS == 1)
    if (ghsMsmt->flagsAvas != 0)
    {
        unsigned short j;
        groupLength++; // for the number of AVAs entry
        for (j = 0; j < ghsMsmt->currentAvaCount; j++)
        {
            groupLength = groupLength + ghsMsmt->avas[j]->length + 6;  // 4-byte id and 2-byte length + 1 byte # of AVAs
        }
    }
    #endif
    return groupLength + 10; // type, length, flags, id
}

static unsigned short computeLengthOfMsmtGroup(s_MsmtGroup *msmtGroup, bool cont)
{
    unsigned short j;
    unsigned short groupLength;
    if (cont)
    {
        groupLength = ((msmtGroup->header->flagsTimeStamp != 0) ? 15 : 7);  // command, flags, length, [timestamp], group id
    }
    else
    {
        groupLength = ((msmtGroup->header->flagsTimeStamp != 0) ? 16 : 8) +
            ((msmtGroup->header->flagsPersonId != 0) ? 2 : 0) +
            ((msmtGroup->header->flagsDuration != 0) ? 4 : 0) +
            ((msmtGroup->header->flagsSuppTypes != 0) ? (1 + msmtGroup->header->numberOfSuppTypes * 4) : 0) +
            ((msmtGroup->header->flagsSrcHndlRef != 0) ? (1 + msmtGroup->header->numberOfSrcHandleRefs * 2) : 0);

        if (msmtGroup->header->flagsAvas != 0)
        {
            groupLength++; // for the number of AVAs entry
            for (j = 0; j < msmtGroup->header->currentAvaCount; j++)
            {
                groupLength = groupLength + msmtGroup->header->avas[j]->length + 6 + 1;
            }
        }
    }
    // Find the length of all GhsMeasurements
    for (j = 0; j < msmtGroup->header->currentMsmtCount; j++)
    {
        s_GhsMsmt* ghs = msmtGroup->ghsMsmts[j];
        groupLength = groupLength + (cont ? 2 : sizeOfGhsBase(msmtGroup->ghsMsmts[j])); // Only the msmt id for the continuous reduced case
        if (ghs->simpleNumeric != NULL)
        {
            groupLength = groupLength + (cont ? 0 : 2) + // units for non-continuous case
                (((ghs->flagsSfloat & MSMT_FLAGS_SFLOAT_VAL) == MSMT_FLAGS_SFLOAT_VAL) ? 2 : 4);
        }
        else if (ghs->compoundNumeric != NULL)
        {
            if (cont)
            {
                groupLength = groupLength + ghs->compoundNumeric->numberOfComponents * ((ghs->flagsSfloat) ? 2 : 4); // sub values
            }
            else
            {
                groupLength = groupLength + 3 + // units, number of components
                    ghs->compoundNumeric->numberOfComponents * ((ghs->flagsSfloat) ? 6 : 8); // sub types + sub values
            }
        }
        else if (ghs->codedEnum != NULL)
        {
            groupLength = groupLength + 4;  // coded enum value
        }
        else if (ghs->bits16Enum != NULL)
        {
            groupLength = groupLength + cont ? 2 : 6; // bits16 enum value set
        }
        else if (ghs->bits32Enum != NULL)
        {
            groupLength = groupLength + cont ? 4 : 12; // bits32 enum value set
        }
        else if (ghs->rtsa != NULL)
        {
            groupLength = groupLength + (cont ? 2 : 17) + ((ghs->rtsa->sampleSize * ghs->rtsa->numberOfSamples) >> 3);
        }
    }
    return groupLength;
}

bool createMsmtGroupDataArray(s_MsmtGroupData** msmtGroupDataPtr, s_MsmtGroup *msmtGroup, unsigned short packetType)
{
    if (msmtGroup == NULL)
    {
        NRF_LOG_DEBUG("MsmtGroup not initialized.");
        return false;
    }
    bool cont = (packetType == PACKET_TYPE_CONT_FOLLOWS);
    unsigned short groupLength = computeLengthOfMsmtGroup(msmtGroup, cont);

    unsigned char *msmtBuf = (unsigned char*)calloc(1, groupLength);
    if (msmtBuf == NULL)
    {
        NRF_LOG_DEBUG("Could not allocate memory for group measurement buffer");
        return false;
    }
    s_MsmtGroupData* msmtGroupData = *msmtGroupDataPtr;
    if (msmtGroupData != NULL)
    {
        cleanUpMsmtGroupData(msmtGroupDataPtr);
    }
    msmtGroupData = (s_MsmtGroupData *)calloc(1, sizeof(s_MsmtGroupData));
    if (msmtGroupData == NULL)
    {
        NRF_LOG_DEBUG("Could not allocate memory for group measurement data");
        free(msmtBuf);
        return false;
    }
    if (msmtGroup->header->currentMsmtCount > 0)
    {
        msmtGroupData->ghsMsmtIndex = (s_GhsMsmtIndex**)calloc(1, msmtGroup->header->currentMsmtCount * sizeof(s_GhsMsmtIndex*));
        if (msmtGroupData->ghsMsmtIndex == NULL)
        {
            NRF_LOG_DEBUG("Could not allocate memory for group measurement index pointers");
            free(msmtBuf);
            free(msmtGroupData);
            return false;
        }
    }
    msmtGroupData->dataLength = (unsigned short)groupLength;
    msmtGroupData->currentGhsMsmtCount = msmtGroup->header->currentMsmtCount;
    
    unsigned short index = 2;  // The first two bytes are reserved for the command
    unsigned short j;

    // Encode the first measurement group
    // header - command

    // header - flags
    unsigned short flags = (unsigned char)(msmtGroup->header->flagsTimeStamp | msmtGroup->header->flagsDuration |
        msmtGroup->header->flagsSuppTypes | msmtGroup->header->flagsSrcHndlRef |
        msmtGroup->header->flagsPersonId | msmtGroup->header->flagsSetting |
        msmtGroup->header->flagsAvas);
    if (cont)
    {
        flags = flags | HEADER_FLAGS_CONT_FOLLOWS;
    }
    else if (packetType == PACKET_TYPE_CONT_FIRST)
    {
        flags = flags | HEADER_FLAGS_CONT_FIRST;
    }
    index = twoByteEncode(msmtBuf, index, flags);

    // header - length of PDU is given by the group length - 6
    index = twoByteEncode(msmtBuf, index, (groupLength - 6));

    // header- timestamp if it exists
    if (msmtGroup->header->flagsTimeStamp != 0)
    {
        index = index + 8;
    }
    if (!cont)
    {
        if (msmtGroup->header->flagsSuppTypes != 0)
        {
            unsigned short i;
            msmtBuf[index++] = (msmtGroup->header->numberOfSuppTypes & 0xFF);
            for (i = 0; i < msmtGroup->header->numberOfSuppTypes; i++)
            {
                index = fourByteEncode(msmtBuf, index, msmtGroup->header->suppTypes[i]);
            }
        }
        if (msmtGroup->header->flagsSrcHndlRef != 0)
        {
            unsigned short i;
            msmtBuf[index++] = (msmtGroup->header->numberOfSrcHandleRefs & 0xFF);
            for (i = 0; i < msmtGroup->header->numberOfSrcHandleRefs; i++)
            {
                index = twoByteEncode(msmtBuf, index, msmtGroup->header->srcHandleRefs[i]);
            }
        }
        if (msmtGroup->header->flagsDuration != 0)
        {
            unsigned long mder;
            createIeeeFloatFromMderFloat(msmtGroup->header->duration, &mder);
            index = fourByteEncode(msmtBuf, index, mder);
        }
        if (msmtGroup->header->flagsPersonId != 0)
        {
            index = twoByteEncode(msmtBuf, index, msmtGroup->header->personId);
        }
        #if (USES_HEADER_AVAS == 1)
        if (msmtGroup->header->flagsAvas != 0)
        {
            int i;
            msmtBuf[index++] = (unsigned char)(unsigned char)msmtGroup->header->currentAvaCount;
            for (i = 0; i < msmtGroup->header->currentAvaCount; i++)
            {
                memcpy(&msmtBuf[index], msmtGroup->header->avas[i], (size_t)(8L + msmtGroup->header->avas[i]->length));
                index = index + 8 + msmtGroup->header->avas[i]->length; // Using a 4-byte id field in AVA
            }
        }
        #endif
    }
    // group id
    msmtBuf[index++] = msmtGroup->header->groupId;

    // number of measurements
    if (!cont) msmtBuf[index++] = (unsigned char)(msmtGroup->header->currentMsmtCount);

    // ================================ Loop: over # of ghs msmts
    for (j = 0; j < msmtGroupData->currentGhsMsmtCount; j++)
    {
        s_GhsMsmtIndex *ghsMsmtIndex = (s_GhsMsmtIndex *)calloc(1, sizeof(s_GhsMsmtIndex));
        if (ghsMsmtIndex == NULL)
        {
            NRF_LOG_DEBUG("Could not allocate memory for group measurement index %d", j);
            free(msmtBuf);
            cleanUpMsmtGroupData(&msmtGroupData);
            return false;
        }
        unsigned short ghsLength = cont ? 2 : 10; // type, length, flags, and id
        s_GhsMsmt* ghs = msmtGroup->ghsMsmts[j];
        // Need to get it now as we don't know the length yet.
        int lengthIndex = index + (cont ? 0 : 4);
        if (cont)
        {
            ghsMsmtIndex->id_index = index;
            index = twoByteEncode(msmtBuf, index, ghs->id);
        }
        else
        {
            index = encodeAlwaysGhsBase(index, ghs, msmtBuf, &ghsMsmtIndex);
        }
        
        //================================= Simple Numeric
        if (ghs->simpleNumeric != NULL)
        {
            #if (USES_NUMERIC == 1)
            ghsMsmtIndex->metricType = MSMT_FLAGS_SIMPLE_NUMERIC;
            if (!cont)
            {
                ghsLength = ghsLength + 2;
                index = twoByteEncode(msmtBuf, index, ghs->simpleNumeric->units);
            }
            // value 
            ghsMsmtIndex->value_index = index;
            ghsMsmtIndex->numberOfValues = 1;
            ghsMsmtIndex->isSfloat = (ghs->flagsSfloat != 0);
            if (ghs->flagsSfloat == 0)
            {
                if(ghs->simpleNumeric->value.mderFloatType == MDER_EMPTY)
                {
                    index = index + 4; // Reserve spot
                }
                else
                {
                    unsigned long ieeeFloat;
                    createIeeeFloatFromMderFloat(&ghs->simpleNumeric->value, &ieeeFloat);
                    index = fourByteEncode(msmtBuf, index, ieeeFloat);
                }
                ghsLength = ghsLength + 4;
            }
            else
            {
                if(ghs->simpleNumeric->value.mderFloatType == MDER_SFLOAT)
                {
                    unsigned short ieeeFloat;
                    createIeeeSFloatFromMderFloat(&ghs->simpleNumeric->value, &ieeeFloat);
                    index = twoByteEncode(msmtBuf, index, ieeeFloat);
                }
                else
                {
                    index = index + 2;
                }
                ghsLength = ghsLength + 2;
            }
            #endif
        }
        //================================= Compound Numeric
        else if (ghs->compoundNumeric != NULL)
        {
            #if (USES_COMPOUND == 1)
            ghsMsmtIndex->metricType = MSMT_FLAGS_COMPOUND_NUMERIC;
            ghsLength = ghsLength + (cont ? 0 : 3);  // units + # of components
            if (!cont)
            {
                // units
                index = twoByteEncode(msmtBuf, index, ghs->compoundNumeric->units);
                // number of components 
                msmtBuf[index++] = (unsigned char)((ghs->compoundNumeric->numberOfComponents) & 0xFF);
            }
            ghsMsmtIndex->value_index = cont ? index : index + 4; // The first value in the compound starts after the 4-byte nomenclature code
            ghsMsmtIndex->numberOfValues = ghs->compoundNumeric->numberOfComponents;
            int k;
            ghsMsmtIndex->isSfloat = (ghs->flagsSfloat != 0);
            for (k = 0; k < ghs->compoundNumeric->numberOfComponents; k++)
            {
                if (!cont)
                {
                    index = fourByteEncode(msmtBuf, index, ghs->compoundNumeric->value[k]->subType);
                    ghsLength = ghsLength + 4;
                }
                if (ghs->flagsSfloat == 0)
                {
                    if(ghs->compoundNumeric->value[k]->subValue.mderFloatType == MDER_EMPTY)
                    {
                        index = index + 4; // Reserve spot
                    }
                    else
                    {
                        unsigned long ieeeFloat;
                        createIeeeFloatFromMderFloat(&ghs->compoundNumeric->value[k]->subValue, &ieeeFloat);
                        index = fourByteEncode(msmtBuf, index, ieeeFloat);
                    }
                    ghsLength = ghsLength + 4;
                }
                else
                {
                    if(ghs->compoundNumeric->value[k]->subValue.mderFloatType == MDER_SFLOAT)
                    {
                        unsigned short ieeeFloat;
                        createIeeeSFloatFromMderFloat(&ghs->compoundNumeric->value[k]->subValue, &ieeeFloat);
                        index = twoByteEncode(msmtBuf, index, ieeeFloat);
                    }
                    else
                    {
                        index = index + 2;
                    }
                    ghsLength = ghsLength + 2;
                }
            }
            #endif
        }

        //================================= Coded Enum
        else if (ghs->codedEnum != NULL)
        {
            #if (USES_CODED == 1)
            ghsMsmtIndex->metricType = MSMT_FLAGS_CODED_ENUM;
            ghsLength = ghsLength + 4;
            // value
            ghsMsmtIndex->value_index = index;
            ghsMsmtIndex->numberOfValues = 1;
            
            index = fourByteEncode(msmtBuf, index, ghs->codedEnum->code);
            #endif
        }

        //================================= Bits 16 Enum
        else if (ghs->bits16Enum != NULL)
        {
            #if (USES_BITS16 == 1)
            ghsMsmtIndex->metricType = MSMT_FLAGS_16BITS_ENUM;
            ghsLength = ghsLength + 6;

            // value 
            ghsMsmtIndex->value_index = index;
            ghsMsmtIndex->numberOfValues = 1;
            // bits
            index = twoByteEncode(msmtBuf, index, ghs->bits16Enum->bitsEnum);
            if (!cont)
            {
                // state
                index = twoByteEncode(msmtBuf, index, ghs->bits16Enum->stateEvent);

                // support 
                index = twoByteEncode(msmtBuf, index, ghs->bits16Enum->supportEvent);
            }
            #endif
        }

        //================================= Bits 32 Enum
        else if (ghs->bits32Enum != NULL)
        {
            #if (USES_BITS32 == 1)
            ghsMsmtIndex->metricType = MSMT_FLAGS_32BITS_ENUM;
            ghsLength = ghsLength + 12;

            // value 
            ghsMsmtIndex->value_index = index;
            ghsMsmtIndex->numberOfValues = 1;
            index = fourByteEncode(msmtBuf, index, ghs->bits32Enum->bitsEnum);
            if (!cont)
            {
                // state 
                index = fourByteEncode(msmtBuf, index, ghs->bits32Enum->stateEvent);
                // support 
                index = fourByteEncode(msmtBuf, index, ghs->bits32Enum->supportEvent);
            }
            #endif
        }
        //================================= RTSA
        else if (ghs->rtsa != NULL)
        {
            #if (USES_RTSA == 1)
            ghsMsmtIndex->metricType = MSMT_FLAGS_RTSA;
            unsigned short rtsaSampleLength = (ghs->rtsa->sampleSize >> 3) * ghs->rtsa->numberOfSamples;
            // units, period, scaleFactor, offset, sampleSize, numberOfSamples, data
            if (!cont)
            {
                // units
                index = twoByteEncode(msmtBuf, index, ghs->rtsa->units);
                // period
                unsigned long mder;
                createIeeeFloatFromMderFloat(&ghs->rtsa->period, &mder);
                index = fourByteEncode(msmtBuf, index, mder);
                // scaleFactor
                createIeeeFloatFromMderFloat(&ghs->rtsa->scaleFactor, &mder);
                index = fourByteEncode(msmtBuf, index, mder);
                // offset
                createIeeeFloatFromMderFloat(&ghs->rtsa->offset, &mder);
                index = fourByteEncode(msmtBuf, index, mder);
                // size
                msmtBuf[index++] = (unsigned char)ghs->rtsa->sampleSize;
            }
            // numberOfSamples
            index = twoByteEncode(msmtBuf, index, ghs->rtsa->numberOfSamples);
            // samples
            ghsMsmtIndex->value_index = index;
            ghsMsmtIndex->numberOfValues = 1;
            if (ghs->rtsa->samples != NULL)
            {
                memcpy(&msmtBuf[index], ghs->rtsa->samples, rtsaSampleLength);
            }
            index = index + rtsaSampleLength;
            ghsLength = ghsLength + 17 + rtsaSampleLength;
            #endif
        }
        msmtGroupData->ghsMsmtIndex[j] = ghsMsmtIndex;
        if (!cont)
        {
            index = encodeOptionals(index, ghs, msmtBuf, &ghsLength);
            twoByteEncode(msmtBuf, lengthIndex, (ghsLength - 6));  // Don't need the updated lengthIndex
        }
    }
    msmtGroupData->data = msmtBuf;
    *msmtGroupDataPtr = msmtGroupData;
    return true;
}
